/* Copyright (C) 2007 Brian Aker, TangentOrg

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#include "mysql_priv.h"
#include "ha_skeleton.h"
#include <mysql/plugin.h>

static handler *skeleton_create_handler(handlerton *hton,
                                       TABLE_SHARE *table, 
                                       MEM_ROOT *mem_root);
static int skeleton_init_func();

/* Variables for skeleton share methods */
static HASH skeleton_open_tables; ///< Hash used to track the number of open tables; variable for skeleton share methods
pthread_mutex_t skeleton_mutex;   ///< This is the mutex used to init the hash; variable for skeleton share methods

static byte* skeleton_get_key(SKELETON_SHARE *share,uint *length,
                             my_bool not_used __attribute__((unused)))
{
  *length=share->table_name_length;
  return (byte*) share->table_name;
}

static int skeleton_init_func(void *p)
{
  DBUG_ENTER("skeleton_init_func");

  handlerton *skeleton_hton;

  skeleton_hton= (handlerton *)p;
  VOID(pthread_mutex_init(&skeleton_mutex,MY_MUTEX_INIT_FAST));
  (void) hash_init(&skeleton_open_tables,system_charset_info,32,0,0,
                   (hash_get_key) skeleton_get_key,0,0);

  skeleton_hton->state=   SHOW_OPTION_YES;
  skeleton_hton->create=  skeleton_create_handler;
  skeleton_hton->flags=   HTON_CAN_RECREATE;

  DBUG_RETURN(0);
}

static int skeleton_done_func(void *p)
{
  int error= 0;
  DBUG_ENTER("skeleton_done_func");

  if (skeleton_open_tables.records)
    error= 1;
  hash_free(&skeleton_open_tables);
  pthread_mutex_destroy(&skeleton_mutex);

  DBUG_RETURN(0);
}

/** @brief
  Skeleton of simple lock controls. The "share" it creates is a structure we will
  pass to each skeleton handler. Do you have to have one of these? Well, you have
  pieces that are used for locking, and they are needed to function.
*/
static SKELETON_SHARE *get_share(const char *table_name, TABLE *table)
{
  SKELETON_SHARE *share;
  uint length;
  char *tmp_name;

  pthread_mutex_lock(&skeleton_mutex);
  length=(uint) strlen(table_name);

  if (!(share=(SKELETON_SHARE*) hash_search(&skeleton_open_tables,
                                           (byte*) table_name,
                                           length)))
  {
    if (!(share=(SKELETON_SHARE *)
          my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
                          &share, sizeof(*share),
                          &tmp_name, length+1,
                          NullS)))
    {
      pthread_mutex_unlock(&skeleton_mutex);
      return NULL;
    }

    share->use_count=0;
    share->table_name_length=length;
    share->table_name=tmp_name;
    strmov(share->table_name,table_name);
    if (my_hash_insert(&skeleton_open_tables, (byte*) share))
      goto error;
    thr_lock_init(&share->lock);
    pthread_mutex_init(&share->mutex,MY_MUTEX_INIT_FAST);
  }
  share->use_count++;
  pthread_mutex_unlock(&skeleton_mutex);

  return share;

error:
  pthread_mutex_destroy(&share->mutex);
  my_free((gptr) share, MYF(0));

  return NULL;
}

/** @brief
  Free lock controls. We call this whenever we close a table. If the table had
  the last reference to the share, then we free memory associated with it.
*/
static int free_share(SKELETON_SHARE *share)
{
  pthread_mutex_lock(&skeleton_mutex);
  if (!--share->use_count)
  {
    hash_delete(&skeleton_open_tables, (byte*) share);
    thr_lock_delete(&share->lock);
    pthread_mutex_destroy(&share->mutex);
    my_free((gptr) share, MYF(0));
  }
  pthread_mutex_unlock(&skeleton_mutex);

  return 0;
}

static handler* skeleton_create_handler(handlerton *hton,
                                       TABLE_SHARE *table, 
                                       MEM_ROOT *mem_root)
{
  return new (mem_root) ha_skeleton(hton, table);
}

ha_skeleton::ha_skeleton(handlerton *hton, TABLE_SHARE *table_arg)
  :handler(hton, table_arg)
{}

static const char *ha_skeleton_exts[] = {
  NullS
};

const char **ha_skeleton::bas_ext() const
{
  return ha_skeleton_exts;
}

int ha_skeleton::open(const char *name, int mode, uint test_if_locked)
{
  DBUG_ENTER("ha_skeleton::open");

  if (!(share = get_share(name, table)))
    DBUG_RETURN(1);
  thr_lock_data_init(&share->lock,&lock,NULL);

  DBUG_RETURN(0);
}

int ha_skeleton::close(void)
{
  DBUG_ENTER("ha_skeleton::close");
  DBUG_RETURN(free_share(share));
}

int ha_skeleton::rnd_init(bool scan)
{
  DBUG_ENTER("ha_skeleton::rnd_init");
  DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

int ha_skeleton::rnd_next(byte *buf)
{
  DBUG_ENTER("ha_skeleton::rnd_next");
  DBUG_RETURN(HA_ERR_END_OF_FILE);
}

void ha_skeleton::position(const byte *record)
{
  DBUG_ENTER("ha_skeleton::position");
  DBUG_VOID_RETURN;
}

int ha_skeleton::rnd_pos(byte * buf, byte *pos)
{
  DBUG_ENTER("ha_skeleton::rnd_pos");
  DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

int ha_skeleton::info(uint flag)
{
  DBUG_ENTER("ha_skeleton::info");
  DBUG_RETURN(0);
}

THR_LOCK_DATA **ha_skeleton::store_lock(THD *thd,
                                       THR_LOCK_DATA **to,
                                       enum thr_lock_type lock_type)
{
  if (lock_type != TL_IGNORE && lock.type == TL_UNLOCK)
    lock.type=lock_type;
  *to++= &lock;
  return to;
}

int ha_skeleton::create(const char *name, TABLE *table_arg,
                       HA_CREATE_INFO *create_info)
{
  DBUG_ENTER("ha_skeleton::create");
  /* This is not implemented but we want someone to be able to see that it works. */
  DBUG_RETURN(0);
}

struct st_mysql_storage_engine skeleton_storage_engine=
{ MYSQL_HANDLERTON_INTERFACE_VERSION };

mysql_declare_plugin(skeleton)
{
  MYSQL_STORAGE_ENGINE_PLUGIN,
  &skeleton_storage_engine,
  "SKELETON",
  "Brian Aker, MySQL AB",
  "Skeleton storage engine",
  PLUGIN_LICENSE_GPL,
  skeleton_init_func, /* Plugin Init */
  skeleton_done_func, /* Plugin Deinit */
  0x0001 /* 0.1 */,
  NULL,                       /* status variables                */
  NULL,                       /* system variables                */
  NULL                        /* config options                  */
}
mysql_declare_plugin_end;
